from typing import Optional
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.storage.provider import StorageProvider
from hub.core.storage.s3 import S3Provider


class LinkCreds(HubMemoryObject):
    def __init__(self):
        self.creds_keys = []
        self.creds_dict = {}  # keys to actual creds dictionary
        self.creds_mapping = {}  # keys to numbers, for encoding
        self.storage_providers = {}
        self.default_s3_provider = None
        self.default_gcs_provider = None

    def get_default_provider(self, provider_type):
        if provider_type == "s3":
            if self.default_s3_provider is None:
                self.default_s3_provider = S3Provider("s3://bucket/path")
            return self.default_s3_provider
        else:
            if self.default_gcs_provider is None:
                from hub.core.storage.gcs import GCSProvider

                self.default_gcs_provider = GCSProvider("gcs://bucket/path")
            return self.default_gcs_provider

    def get_storage_provider(self, key: Optional[str], provider_type):
        assert provider_type in {"s3", "gcs"}
        if key in {"ENV", None}:
            return self.get_default_provider(provider_type)
        if key not in self.creds_keys:
            raise KeyError(f"Creds key {key} does not exist")
        if key not in self.creds_dict:
            raise ValueError(
                f"Creds key {key} hasn't been populated. Populate it using ds.populate_creds()"
            )

        provider: StorageProvider
        creds = self.creds_dict[key]

        if provider_type == "s3":
            if key in self.storage_providers:
                provider = self.storage_providers[key]
                if isinstance(provider, S3Provider):
                    return provider

            provider = S3Provider("s3://bucket/path", **creds)
        else:
            from hub.core.storage.gcs import GCSProvider

            if key in self.storage_providers:
                provider = self.storage_providers[key]
                if isinstance(provider, GCSProvider):
                    return provider

            provider = GCSProvider("gcs://bucket/path", **creds)
        self.storage_providers[key] = provider
        return provider

    def add_creds(self, creds_key: str):
        if creds_key in self.creds_keys:
            raise ValueError(f"Creds key {creds_key} already exists")
        self.creds_keys.append(creds_key)
        self.creds_mapping[creds_key] = len(self.creds_keys)

    def populate_creds(self, creds_key: str, creds):
        if creds_key not in self.creds_keys:
            raise KeyError(f"Creds key {creds_key} does not exist")
        self.creds_dict[creds_key] = creds

    def tobytes(self) -> bytes:
        return bytes(",".join(self.creds_keys), "utf-8")

    def get_encoding(self, key):
        if key in {None, "ENV"}:
            return 0
        if key not in self.creds_keys:
            raise ValueError(f"Creds key {key} does not exist")
        return self.creds_mapping[key]

    def get_creds_key(self, encoding):
        if encoding > len(self.creds_keys):
            raise KeyError(f"Encoding {encoding} not found.")
        return None if encoding == 0 else self.creds_keys[encoding - 1]

    @classmethod
    def frombuffer(cls, buffer: bytes):
        obj = cls()
        if buffer:
            obj.creds_keys = list(buffer.decode("utf-8").split(","))
            obj.creds_mapping = {k: i + 1 for i, k in enumerate(obj.creds_keys)}
        obj.is_dirty = False
        return obj

    @property
    def nbytes(self):
        return len(self.tobytes())

    def __getstate__(self):
        return {
            "creds_keys": self.creds_keys,
            "creds_dict": self.creds_dict,
        }

    def __setstate__(self, state):
        self.creds_keys = state["creds_keys"]
        self.creds_dict = state["creds_dict"]
        self.creds_mapping = {key: i + 1 for i, key in enumerate(self.creds_keys)}
        self.storage_providers = {}
        self.default_s3_provider = None
        self.default_gcs_provider = None

    def __len__(self):
        return len(self.creds_keys)

    @property
    def missing_keys(self) -> list:
        return [key for key in self.creds_keys if key not in self.creds_dict]



from hub.core.linked_sample import LinkedSample
from typing import Optional, Dict


def link(
    path: str,
    creds_key: Optional[str] = None,
) -> LinkedSample:
    """Utility that stores a link to raw data. Used to add data to a Hub Dataset without copying it.

    Note:
        No data is actually loaded until you try to read the sample from a dataset.
        There are a few exceptions to this:-

        - If verify=True was specified DURING create_tensor of the tensor to which this is being added, some metadata is read to verify the integrity of the sample.
        - If create_shape_tensor=True was specified DURING create_tensor of the tensor to which this is being added, the shape of the sample is read.
        - If create_sample_info_tensor=True was specified DURING create_tensor of the tensor to which this is being added, the sample info is read.

    Examples:
        >>> ds = hub.dataset("......")

        Add the names of the creds you want to use (not needed for http/local urls)
        >>> ds.add_creds("MY_S3_KEY")
        >>> ds.add_creds("GCS_KEY")

        Populate the names added with creds dictionary
        These creds are only present temporarily and will have to be repopulated on every reload
        >>> ds.populate_creds("MY_S3_KEY", {})
        >>> ds.populate_creds("GCS_KEY", {})

        Create a tensor that can contain links
        >>> ds.create_tensor("img", htype="link[image]", verify=True, create_shape_tensor=False, create_sample_info_tensor=False)

        Populate the tensor with links
        >>> ds.img.append(hub.link("s3://abc/def.jpeg", creds_key="MY_S3_KEY"))
        >>> ds.img.append(hub.link("gcs://ghi/jkl.png", creds_key="GCS_KEY"))
        >>> ds.img.append(hub.link("https://picsum.photos/200/300")) # doesn't need creds
        >>> ds.img.append(hub.link("s3://abc/def.jpeg"))  # will use creds from environment
        >>> ds.img.append(hub.link("s3://abc/def.jpeg", creds_key="ENV"))  # this will also use creds from environment

        Accessing the data
        >>> for i in range(5):
        >>> ds.img[i].numpy()

        Updating a sample
        >>> ds.img[0] = hub.link("./data/cat.jpeg")

    Supported file types:

        Image: "bmp", "dib", "gif", "ico", "jpeg", "jpeg2000", "pcx", "png", "ppm", "sgi", "tga", "tiff", "webp", "wmf", "xbm"
        Audio: "flac", "mp3", "wav"
        Video: "mp4", "mkv", "avi"
        Dicom: "dcm"

    Args:
        path (str): Path to a supported file.
        creds_key (optional, str): The credential key to use to read data for this sample. The actual credentials are fetched from the dataset.

    Returns:
        LinkedSample: LinkedSample object that stores path and creds.
    """
    return LinkedSample(path, creds_key)



from hub.core.link_creds import LinkCreds
from hub.core.lock import Lock
from hub.core.storage.lru_cache import LRUCache
from hub.util.keys import (
    get_dataset_linked_creds_key,
    get_dataset_linked_creds_lock_key,
)
from hub.util.remove_cache import get_base_storage


def merge_link_creds(old_link_creds: LinkCreds, current_link_creds: LinkCreds):
    num_common_keys = 0
    for key1, key2 in zip(old_link_creds.creds_keys, current_link_creds.creds_keys):
        if key1 == key2:
            num_common_keys += 1
        else:
            break
    new_keys = current_link_creds.creds_keys[num_common_keys:]
    current_link_creds.creds_keys = old_link_creds.creds_keys
    current_link_creds.creds_mapping = old_link_creds.creds_mapping
    for key in new_keys:
        if key not in current_link_creds.creds_mapping:
            current_link_creds.add_creds(key)
    return current_link_creds


def save_link_creds(current_link_creds: LinkCreds, storage: LRUCache):
    """Saves the linked creds info to storage."""
    storage = get_base_storage(storage)
    lock = Lock(storage, get_dataset_linked_creds_lock_key())
    lock.acquire(timeout=10, force=True)
    key = get_dataset_linked_creds_key()
    try:
        data_bytes = storage[key]
    except KeyError:
        data_bytes = None

    if data_bytes is not None:
        old_link_creds = LinkCreds.frombuffer(data_bytes)
        new_link_creds = merge_link_creds(old_link_creds, current_link_creds)
    else:
        new_link_creds = current_link_creds

    storage[key] = new_link_creds.tobytes()
    lock.release()



from typing import Optional, Dict, Any, Tuple
import hub
from hub.core.chunk_engine import ChunkEngine
from hub.core.index.index import Index
from hub.core.link_creds import LinkCreds
from hub.core.linked_sample import LinkedSample
from hub.core.meta.encode.creds import CredsEncoder
from hub.core.storage import LRUCache
from hub.core.tensor_link import read_linked_sample
from hub.core.compression import _read_video_shape, _decompress_video
from hub.util.exceptions import ReadOnlyModeError
from hub.util.keys import get_creds_encoder_key
from hub.util.video import normalize_index


class LinkedChunkEngine(ChunkEngine):
    def __init__(
        self,
        key: str,
        cache: LRUCache,
        version_state: Dict[str, Any],
        link_creds: LinkCreds,
        meta_cache: LRUCache = None,
    ):
        super().__init__(key, cache, version_state, meta_cache)
        self.link_creds = link_creds
        self._creds_encoder: Optional[CredsEncoder] = None
        self._creds_encoder_commit_id: Optional[str] = None

    @property
    def creds_encoder(self):
        commit_id = self.commit_id
        if self._creds_encoder is None or self._creds_encoder_commit_id != commit_id:
            commit_id = self.commit_id
            key = get_creds_encoder_key(self.key, commit_id)
            if not self.creds_encoder_exists:
                enc = CredsEncoder()
                try:
                    self.meta_cache[key] = enc
                except ReadOnlyModeError:
                    pass
            else:
                enc = self.meta_cache.get_hub_object(key, CredsEncoder)
            self._creds_encoder = enc
            self._creds_encoder_commit_id = commit_id
            self.meta_cache.register_hub_object(key, enc)
        return self._creds_encoder

    @property
    def creds_encoder_exists(self):
        commit_id = self.commit_id
        if (
            self._creds_encoder is not None
            and self._creds_encoder_commit_id == commit_id
        ):
            return True
        try:
            key = get_creds_encoder_key(self.key, commit_id)
            self.meta_cache[key]
            return True
        except KeyError:
            return False

    @property
    def is_data_cachable(self):
        return False

    def linked_sample(self, global_sample_index: int) -> LinkedSample:
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index)
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        return LinkedSample(sample_path, sample_creds_key)

    def get_video_sample(self, global_sample_index, index):
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index)
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        storage = None
        if sample_path.startswith(("gcs://", "gcp://", "s3://")):
            provider_type = "s3" if sample_path.startswith("s3://") else "gcs"
            storage = self.link_creds.get_storage_provider(
                sample_creds_key, provider_type
            )
            url = storage.get_presigned_url(sample_path, full=True)
        else:
            url = sample_path
        squeeze = isinstance(index, int)
        shape = _read_video_shape(url)
        sub_index = index.values[1].value if len(index.values) > 1 else None  # type: ignore
        start, stop, step, reverse = normalize_index(sub_index, shape[0])
        video_sample = _decompress_video(
            url,
            start,
            stop,
            step,
            reverse,
        )
        if squeeze:
            video_sample.squeeze(0)
        return video_sample

    def get_basic_sample(self, global_sample_index, index):
        sample = self.get_hub_read_sample(global_sample_index)
        return sample.array[tuple(entry.value for entry in index.values[1:])]

    def get_path(self, global_sample_index) -> str:
        return super().get_basic_sample(
            global_sample_index, Index(global_sample_index)
        )[0]

    def get_hub_read_sample(self, global_sample_index):
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index)
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        return read_linked_sample(sample_path, sample_creds_key, self.link_creds, False)

    @property
    def verify(self):
        return self.tensor_meta.is_link and self.tensor_meta.verify

    def check_each_sample(self, samples):
        verified_samples = []
        for i, sample in enumerate(samples):
            if isinstance(sample, hub.core.tensor.Tensor) and sample.is_link:
                sample = sample._linked_sample()
                samples[i] = sample
            elif not isinstance(sample, LinkedSample) and sample is not None:
                raise TypeError(
                    f"Expected LinkedSample, got {type(sample)} instead. Use hub.link() to link samples."
                )
            if self.verify:
                if sample is None:
                    verified_samples.append(sample)
                else:
                    verified_samples.append(
                        read_linked_sample(
                            sample.path, sample.creds_key, self.link_creds, verify=True
                        )
                    )
        return verified_samples

    def register_new_creds(self, num_samples_added, samples):
        assert isinstance(num_samples_added, int)
        link_creds = self.link_creds
        creds_encoder = self.creds_encoder
        for i in range(num_samples_added):
            creds_key = samples[i].creds_key
            encoded_creds_key = link_creds.get_encoding(creds_key)
            creds_encoder.register_samples((encoded_creds_key,), 1)

    def update_creds(self, sample_index: int, sample: LinkedSample):
        link_creds = self.link_creds
        creds_key = sample.creds_key
        encoded_creds_key = link_creds.get_encoding(creds_key)
        self.creds_encoder[sample_index] = (encoded_creds_key,)

    def read_shape_for_sample(self, global_sample_index: int) -> Tuple[int, ...]:
        sample = self.get_hub_read_sample(global_sample_index)
        return sample.shape
